from django.urls import path

from . import views

urlpatterns = [
    path('login/', views.login_view, name='login'),
    path('logout/', views.logout, name='logout'),

    path('sso/test/', views.for_logged_only),
    path('sso/push/', views.accept_user_information),
    path('sso/accept/', views.authorize_from_sso),
    path('sso/deauthenticate/', views.deauthenticate, name="deauthenticate_view"),
]


from importlib import import_module

from django.conf import settings
from django.contrib.auth import get_user_model


def deauthenticate_user(user_identy):
    """
    Deauthenticates user in currently runing context
    """
    from django.contrib.sessions.models import Session

    user_model = get_user_model()

    user = user_model.objects.filter(**{
        f'{user_model.USERNAME_FIELD}': user_identy
    }).first()

    if user:
        for session_key in Session.objects.values_list('session_key', flat=True).all():
            session = import_module(settings.SESSION_ENGINE).SessionStore(session_key)

            if '_auth_user_id' in session and int(session['_auth_user_id']) == user.id:
                del session['_auth_user_id']

                if '_auth_user_backend' in session:
                    del session['_auth_user_backend']

                if '_auth_user_hash' in session:
                    del session['_auth_user_hash']

                session.save()










from django.contrib import admin
from .models import Service, AuthenticationRequest


@admin.register(Service)
class ServiceAdmin(admin.ModelAdmin):
	list_display = 'name', 'base_url', 'enabled'
	fields = 'name', 'base_url', 'token', 'enabled'
	readonly_fields = 'token',


@admin.register(AuthenticationRequest)
class AuthenticationRequestAdmin(admin.ModelAdmin):
	readonly_fields = 'token',
	list_display = (
		'service',
		'user_identy',
		'created_at',
		'used',
	)

	list_filter = 'service', 'used'
	search_fields = 'user_identy',




import requests
from django.conf import settings

from django_sso.exceptions import SSOException


def set_sso_authorization_request_used(sso_token):
    """
    For service side. Makes SSO request as used for authentication procedure (not available for next authentications)
    """
    try:
        result = requests.post(settings.SSO_ROOT.rstrip('/') + '/sso/make_used/', {
            'token': settings.SSO_TOKEN,
            'authentication_token': sso_token
        })

        if result.status_code != 200:
            raise Exception(f'Некорректный ответ сервера авторизации: STATUS={result.status_code}; TEXT={result.text}')
    except Exception as e:
        raise SSOException(e)


# Generated by Django 3.2.7 on 2022-04-14 07:43

from django.db import migrations, models
import django_sso.gate.models


class Migration(migrations.Migration):

    dependencies = [
        ('gate', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='service',
            name='token',
            field=models.CharField(default=django_sso.gate.models.service_token_generator, max_length=128, unique=True, verbose_name='Token'),
        ),
    ]



import urllib.parse

from django.conf import settings
from django.contrib.auth import login, get_user_model
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse, HttpResponse
from django.shortcuts import render, redirect
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
import django.contrib.auth

from django_sso import deauthenticate_user
from django_sso.exceptions import SSOException
from django_sso.gate import set_sso_authorization_request_used
from django_sso.service import get_sso_authorization_request, request_sso_authorization_request, \
    request_deauthentication


def login_view(request):
    """
    The view to redirect to SSO for authentication
    """
    try:
        sso_token = request_sso_authorization_request(request)
        request.session['token'] = sso_token
    except SSOException as e:
        return render(
            request,
            template_name='django_sso/token_validation_error.html',
            context={
                'error': str(e),
                'debug': settings.DEBUG
            }
        )

    params = urllib.parse.urlencode({'sso': sso_token})

    return render(
        request,
        context={
            'redirect_url': f"{settings.SSO_ROOT}/login/?{params}"
        },
        template_name='django_sso/redirect_to_sso.html'
    )


@login_required
def for_logged_only(request):
    """
    The view for testing mechanism
    """
    user_model = get_user_model()
    return HttpResponse(f'{_("You authorized as")} <b>{getattr(request.user, user_model.USERNAME_FIELD)}</b>.')


@csrf_exempt
def accept_user_information(request):
    """
    Создаёт пользователя по команде SSO-сервера. Удостоверение производится при
    помощи токена, который указан в настройках текщего проекта для удостоверения
    на SSO-сервере.
    """
    if (
            request.method != 'POST'
            or 'token' not in request.POST
            or 'username' not in request.POST
            or 'password' not in request.POST
            or request.POST['token'] != settings.SSO_TOKEN
    ):
        return JsonResponse({'error': _('Incorrect request')}, status=400)

    user_model = get_user_model()

    user, created = User.objects.get_or_create(
        **{f'{user_model.USERNAME_FIELD}': request.POST['username']}
    )

    if user.password != request.POST['password']:
        user.password = request.POST['password']
        user.save()

    return JsonResponse({'ok': True})


def authorize_from_sso(request: WSGIRequest):
    """
    Авторизует пользователя, который вернулся от SSO-сервера
    """
    if not request.session.session_key or not request.session.get('token', ''):
        return redirect('/')

    try:
        authorization_request = get_sso_authorization_request(request.session.get('token'))

        user_model = get_user_model()

        user = User.objects.filter(
            **{f'{user_model.USERNAME_FIELD}': authorization_request['user_identy']}
        ).first()

        if not user:
            raise SSOException(_('Local user for SSO does not exist'))

        set_sso_authorization_request_used(request.session.get('token'))

        login(request, user)

        if authorization_request['next_url']:
            return redirect(authorization_request['next_url'])
        else:
            return redirect('/')
    except SSOException as e:
        if settings.DEBUG:
            return render(request, 'django_sso/token_validation_error.html', context={
                'error': e
            })
        else:
            return redirect('/')


def logout(request):
    if not request.user.is_anonymous:
        try:
            request_deauthentication(request.user)
            django.contrib.auth.logout(request)
        except Exception as e:
            if settings.DEBUG:
                return render(request, template_name='django_sso/deauthentication_error.html', context={
                    'error': e
                })

    return redirect('/')


@csrf_exempt
def deauthenticate(request):
    """
    Accept deauthenticate request from SSO server and process it
    """
    if request.method != 'POST':
        return HttpResponse(status=405)

    if (
        not request.POST.get('token', '').strip()
        or not request.POST.get('user_identy', '').strip()
    ):
        return JsonResponse({'error': _('Incorrect request')})

    user_model = get_user_model()

    user = user_model.objects.filter(**{
        f'{user_model.USERNAME_FIELD}': request.POST['user_identy']
    }).first()

    if user:
        deauthenticate_user(getattr(user, user_model.USERNAME_FIELD))

    return JsonResponse({'ok': True})



import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.db import models
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

from django_sso.exceptions import SSOException


def service_token_generator():
	return get_random_string(Service.token.field.max_length)


class Service(models.Model):
	name = models.CharField(max_length=128, verbose_name=_('Name'))
	base_url = models.URLField(verbose_name=_('Base url'))
	enabled = models.BooleanField(default=False, verbose_name=_('Enabled'))
	token = models.CharField(max_length=128, verbose_name=_('Token'), unique=True, default=service_token_generator)

	def __str__(self):
		return self.base_url

	def requeset_deauthenticate(self, user_identy):
		"""
		push deauthentication event to subordinate service
		"""
		if not self.enabled:
			return

		try:
			data = requests.post(f'{self.base_url}/sso/deauthenticate/', {
				'token': self.token,
				'user_identy': user_identy
			})
		except Exception as e:
			if settings.DEBUG:
				raise e

	class Meta:
		verbose_name = _('External service')
		verbose_name_plural = _('External services')


def auth_token_generator():
	return get_random_string(AuthenticationRequest.token.field.max_length)


class AuthenticationRequest(models.Model):
	service = models.ForeignKey('Service', on_delete=models.CASCADE, verbose_name=_('Service'))
	created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
	token = models.CharField(max_length=128, verbose_name=_('Token'), default=auth_token_generator, unique=True)
	user_identy = models.CharField(max_length=128, verbose_name=_('User identy'), help_text=_('E-Mail, login, etc.'))
	next_url = models.CharField(max_length=512, verbose_name=_('Next url'), help_text=_('To go after success auth'))
	authenticated = models.BooleanField(default=False, verbose_name=_('Request has been activated'))
	used = models.BooleanField(default=False, verbose_name=_('Are used in external service'))

	class Meta:
		verbose_name = _('Authentication request')
		verbose_name_plural = _('Authentication requests')

	def activate(self, user: User):
		"""
		1) Активация запроса авторизации
		2) Отправка на подчинённый сервис базовой информации о пользователе для создания,
		если там отсутствует такой аккаунт.
		"""
		user_model = get_user_model()

		self.user_identy = getattr(user, user_model.USERNAME_FIELD)
		self.authenticated = True
		self.save()

		try:
			user_model = get_user_model()

			response = requests.post(f'{self.service.base_url}/sso/push/', {
				'token': self.service.token,
				'username': getattr(user, user_model.USERNAME_FIELD),
				'password': user.password
			}).json()

			if 'ok' in response:
				return
			elif 'error' in response:
				raise SSOException(response['error'])
			else:
				raise NotImplementedError(_('From service returned incorrect data'))
		except Exception as e:
			raise SSOException(str(e))





# Generated by Django 3.2.7 on 2022-04-14 04:19

from django.db import migrations, models
import django.db.models.deletion
import django_sso.gate.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('base_url', models.URLField(verbose_name='Base url')),
                ('enabled', models.BooleanField(default=False, verbose_name='Enabled')),
                ('token', models.CharField(default=django_sso.gate.models.service_token_generator, max_length=256, unique=True, verbose_name='Token')),
            ],
            options={
                'verbose_name': 'External service',
                'verbose_name_plural': 'External services',
            },
        ),
        migrations.CreateModel(
            name='AuthenticationRequest',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('token', models.CharField(default=django_sso.gate.models.auth_token_generator, max_length=128, unique=True, verbose_name='Token')),
                ('user_identy', models.CharField(help_text='E-Mail, login, etc.', max_length=128, verbose_name='User identy')),
                ('used', models.BooleanField(default=False, verbose_name='Are used in external service')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='gate.service', verbose_name='Service')),
            ],
            options={
                'verbose_name': 'Authentication request',
                'verbose_name_plural': 'Authentication requests',
            },
        ),
    ]



from builtins import super
from typing import Optional

import django.contrib.auth.views
from django.contrib.auth import logout, get_user_model
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View
from django.views.generic.edit import FormView

from .models import Service, AuthenticationRequest
from .. import deauthenticate_user
from ..exceptions import SSOException


class LoginView(django.contrib.auth.views.LoginView):
	template_name = 'django_sso/login.html'

	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super(FormView, self).dispatch(request, *args, **kwargs)

	def get_success_url(self):
		sso_request_token = self.request.GET.get('sso', '').strip()
		auth_request = None

		if sso_request_token:
			auth_request: Optional[AuthenticationRequest] = AuthenticationRequest.objects.filter(
				token=sso_request_token,
				used=False,
			).first()

		if not auth_request or not auth_request.next_url:
			return reverse_lazy('welcome')

		try:
			auth_request.activate(self.request.user)
		except SSOException:
			return reverse_lazy('welcome')

		return f'{auth_request.service.base_url}/sso/accept/'

	def form_valid(self, form):
		super().form_valid(form)

		return redirect(self.get_success_url())


class LogoutView(View):
	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super().dispatch(request, *args, **kwargs)

	def post(self, request):
		logout(request)

		return redirect(reverse_lazy('login'))


class ObtainView(View):
	"""
	The view for external services for obtain SSO token
	"""

	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super().dispatch(request, *args, **kwargs)

	def post(self, request):
		if not request.POST.get('token', '').strip():
			return JsonResponse({"error": _('"token" is\'nt set')})

		if not request.POST.get('next_url', '').strip():
			return JsonResponse({"error": _('"next_url" is\'nt set')})

		service = Service.objects.filter(enabled=True, token=request.POST['token']).first()

		if not service:
			return JsonResponse({"error": _("Application token is'nt exist")})

		sso_request = AuthenticationRequest(
			service=service,
			next_url=request.POST['next_url']
		)

		sso_request.save()

		return JsonResponse({'token': sso_request.token})


class GetAuthenticationRequestView(View):
	"""
	The view for the services to do obtain information about the authentication request by SSO-token
	If
	- token exist for requested service
	- token are activated and wasn't used yet for login
	will be returned the authentication request information as is, else 401 code with error
	"""

	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super().dispatch(request, *args, **kwargs)

	def post(self, request):
		if not request.POST.get('token', '').strip():
			return JsonResponse({"error": _('"token" is\'nt set')})

		service = Service.objects.filter(enabled=True, token=request.POST['token']).first()

		if not service:
			return JsonResponse({"error": _("Application token is'nt exist")})

		if not request.POST.get('authentication_token', '').strip():
			return JsonResponse({"error": _("Authentication request token is'nt set")})

		authorization_request = AuthenticationRequest.objects.filter(
			token=request.POST.get('authentication_token'),
			service=service,
			authenticated=True,
			used=False,
		).first()

		if not authorization_request:
			return JsonResponse({"error": _("Authentication request token is'nt exists")})
		else:
			return JsonResponse({
				"authenticated": True,
				"user_identy": authorization_request.user_identy,
				"next_url": authorization_request.next_url
			})


class MakeUsedView(View):
	"""
	The view who mark authentication request as used
	"""

	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super().dispatch(request, *args, **kwargs)

	def post(self, request):
		if not request.POST.get('token', '').strip():
			return JsonResponse({"error": _('"token" is\'nt set')})

		service = Service.objects.filter(enabled=True, token=request.POST['token']).first()

		if not service:
			return JsonResponse({"error": _("Application token is'nt exist")})

		if not request.POST.get('authentication_token', '').strip():
			return JsonResponse({"error": _("Authentication request token is'nt set")})

		authorization_request = AuthenticationRequest.objects.filter(
			token=request.POST.get('authentication_token'),
			service=service,
			authenticated=True,
			used=False,
		).first()

		if not authorization_request:
			return JsonResponse({"error": _("Authentication request token is'nt exists")}, status=404)
		else:
			authorization_request.authenticated = True
			authorization_request.save()

			return JsonResponse({"ok": True})


class DeauthenticateView(View):
	"""
	Deauthenticate user everywhere by identy.
	Deauth on all services except requester
	"""

	@csrf_exempt
	def dispatch(self, request, *args, **kwargs):
		return super().dispatch(request, *args, **kwargs)

	def post(self, request):
		if not request.POST.get('token', '').strip():
			return JsonResponse({"error": _('"token" is\'nt set')})

		if not request.POST.get('user_identy', '').strip():
			return JsonResponse({"error": _('"user_identy" is\'nt set')})

		service = Service.objects.filter(enabled=True, token=request.POST['token']).first()

		user_model = get_user_model()

		user = user_model.objects.filter(**{
			f'{user_model.USERNAME_FIELD}': request.POST['user_identy']
		}).first()

		if user:
			deauthenticate_user(getattr(user, user_model.USERNAME_FIELD))

		for instance in Service.objects.filter(enabled=True).exclude(id=service.id).all():
			instance.requeset_deauthenticate()

		return JsonResponse({'ok': True})


# Generated by Django 3.2.7 on 2022-04-14 11:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gate', '0002_alter_service_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticationrequest',
            name='next_url',
            field=models.CharField(default=None, help_text='To go after success auth', max_length=512, verbose_name='Next url'),
            preserve_default=False,
        ),
    ]



from setuptools import setup, find_packages

setup(
    name='django_sso',
    version='1.0.0',
    author="DAVIDhaker",
    author_email='me@davidhaker.ru',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/davidhaker/django-sso',
    keywords='Django SSO Single Sign-On',
    install_requires=[
        'django',
    ],
)



# Generated by Django 3.2.7 on 2022-04-21 04:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gate', '0003_authenticationrequest_next_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticationrequest',
            name='authenticated',
            field=models.BooleanField(default=False, verbose_name='Request has been activated'),
        ),
    ]
